import os
import unittest
from copy import deepcopy
from tempfile import TemporaryDirectory

from dhpython.depends import Dependencies

from .test_depends import FakeOptions, prime_pydist


class TestDistutilsExtra(unittest.TestCase):
    options = FakeOptions(guess_deps=True)
    pydist = {
        'bar': 'python3-bar',
        'baz': {'dependency': 'python3-baz', 'standard': 'PEP386'},
        'quux': {'dependency': 'python3-quux', 'standard': 'PEP386'},
    }
    pkg = 'foo'
    impl = 'cpython3'
    stats = {
        'compile': False,
        'dist-info': set(),
        'egg-info': set(('PKG-INFO',)),
        'ext_no_version': set(),
        'ext_vers': set(),
        'nsp.txt': set(),
        'private_dirs': {},
        'public_vers': set(),
        'requires.txt': set(),
        'shebangs': set(),
    }
    requires = {}

    def test_depends_on_bar(self):
        self.d = Dependencies(self.pkg, self.impl)
        stats = deepcopy(self.stats)
        self.tempdir = TemporaryDirectory()  # pylint: disable=consider-using-with
        self.addCleanup(self.tempdir.cleanup)
        old_wd = os.getcwd()
        os.chdir(self.tempdir.name)
        self.addCleanup(os.chdir, old_wd)
        with open(self.tempdir.name + '/PKG-INFO', 'w', encoding="UTF-8") as f:
            f.write("""Metadata-Version: 2.1
Name: gTranscribe
Version: 0.11
Summary: gTranscribe
Home-page: https://github.com/innir/gtranscribe
Author: Philip Rinn
Author-email: rinni@inventati.org
License: GPL-3
Requires: bar

gTranscribe is a software focused on easy transcription of spoken words.
""")
        cleanup = prime_pydist(self.impl, self.pydist)
        self.addCleanup(cleanup)
        self.d.parse(stats, self.options)
        self.assertIn('python3-bar', self.d.depends)



from tempfile import TemporaryDirectory
import os
import unittest

from dhpython.tools import relpath, move_matching_files


class TestRelpath(unittest.TestCase):
    def test_common_parent_dir(self):
        r = relpath('/usr/share/python-foo/foo.py', '/usr/bin/foo')
        self.assertEqual(r, '../share/python-foo/foo.py')

    def test_strips_common_prefix(self):
        r = relpath('/usr/share/python-foo/foo.py', '/usr/share')
        self.assertEqual(r, 'python-foo/foo.py')

    def test_trailing_slash_ignored(self):
        r = relpath('/usr/share/python-foo/foo.py', '/usr/share/')
        self.assertEqual(r, 'python-foo/foo.py')


class TestMoveMatchingFiles(unittest.TestCase):
    def setUp(self):
        self.tmpdir = TemporaryDirectory()  # pylint: disable=consider-using-with
        self.addCleanup(self.tmpdir.cleanup)
        os.makedirs(self.tmppath('foo/bar/a/b/c/spam'))
        for path in ('foo/bar/a/b/c/spam/file.so',
                     'foo/bar/a/b/c/spam/file.py'):
            with open(self.tmppath(path), 'wb'):
                # create a 0 byte file for the test
                pass

        move_matching_files(self.tmppath('foo/bar/'),
                            self.tmppath('foo/baz/'),
                            r'spam/.*\.so$')

    def tmppath(self, *path):
        return os.path.join(self.tmpdir.name, *path)

    def test_moved_matching_file(self):
        self.assertTrue(os.path.exists(
            self.tmppath('foo/baz/a/b/c/spam/file.so')))

    def test_left_non_matching_file(self):
        self.assertTrue(os.path.exists(
            self.tmppath('foo/bar/a/b/c/spam/file.py')))



# Copyright © 2022 Stefano Rivera <stefanor@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Handle Environment Markers
https://www.python.org/dev/peps/pep-0508/#environment-markers

TODO: Ideally replace with the packaging library, but the API is currently
private: https://github.com/pypa/packaging/issues/496
"""

import re


SIMPLE_ENV_MARKER_RE = re.compile(r'''
    (?P<marker>[a-z_]+)
    \s*
    (?P<op><=?|>=?|[=!~]=|===)
    \s*
    (?P<quote>['"])
    (?P<value>.*)  # Could contain additional markers
    (?P=quote)
    ''', re.VERBOSE)
COMPLEX_ENV_MARKER_RE = re.compile(r'''
    (?:\s|\))
    (?:and|or)
    (?:\s|\()
    ''', re.VERBOSE)


class ComplexEnvironmentMarker(Exception):
    pass


def parse_environment_marker(marker):
    """Parse a simple marker of <= 1 environment restriction"""
    marker = marker.strip()
    if marker.startswith('(') and marker.endswith(')'):
        marker = marker[1:-1].strip()

    m = COMPLEX_ENV_MARKER_RE.search(marker)
    if m:
        raise ComplexEnvironmentMarker()

    m = SIMPLE_ENV_MARKER_RE.match(marker)
    if not m:
        raise ComplexEnvironmentMarker()

    return (
        m.group('marker'),
        m.group('op'),
        m.group('value'),
    )



import os
from unittest import TestCase
import subprocess


class RequiredTest(TestCase):
    def test_tests_are_executed(self):
        open('test-executed', 'w').close()

    def test_entry_point_executed(self):
        path, _, __ = os.environ['PATH'].partition(":")
        assert path.endswith("/scripts")
        subprocess.run('foo', check=True)



# Copyright © 2012-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
from glob import glob1
from os import remove
from os.path import exists, isdir, join
from shutil import rmtree, move
from dhpython.build.base import Base, shell_command, copy_test_files

log = logging.getLogger('dhpython')
_setup_tpl = 'setup.py|setup-3.py'


def create_pydistutils_cfg(func):
    """distutils doesn't have sane command-line API - this decorator creates
    .pydistutils.cfg file to workaround it

    hint: if you think this is plain stupid, please don't read
    distutils/setuptools/distribute sources
    """

    def wrapped_func(self, context, args, *oargs, **kwargs):
        fpath = join(args['home_dir'], '.pydistutils.cfg')
        if not exists(fpath):
            with open(fpath, 'w', encoding='utf-8') as fp:
                lines = ['[clean]\n',
                         'all=1\n',
                         '[build]\n',
                         'build_lib={}\n'.format(args['build_dir']),
                         '[install]\n',
                         'force=1\n',
                         'install_layout=deb\n',
                         'install_scripts=$base/bin\n',
                         'install_lib={}\n'.format(args['install_dir']),
                         'prefix=/usr\n']
                log.debug('pydistutils config file:\n%s', ''.join(lines))
                fp.writelines(lines)
        context['ENV']['HOME'] = args['home_dir']
        return func(self, context, args, *oargs, **kwargs)

    wrapped_func.__name__ = func.__name__
    return wrapped_func


class BuildSystem(Base):
    DESCRIPTION = 'Distutils build system'
    SUPPORTED_INTERPRETERS = {'python', 'python3', 'python{version}',
                              'python-dbg', 'python3-dbg', 'python{version}-dbg'}
    REQUIRED_FILES = [_setup_tpl]
    OPTIONAL_FILES = {'setup.cfg': 1,
                      'requirements.txt': 1,
                      'PKG-INFO': 10,
                      '*.egg-info': 10}
    CLEAN_FILES = Base.CLEAN_FILES | {'build'}

    def detect(self, context):
        result = super().detect(context)
        if _setup_tpl in self.DETECTED_REQUIRED_FILES:
            context['args']['setup_py'] = self.DETECTED_REQUIRED_FILES[_setup_tpl][0]
        else:
            context['args']['setup_py'] = 'setup.py'
        return result

    @shell_command
    @create_pydistutils_cfg
    def clean(self, context, args):
        super().clean(context, args)
        if exists(args['interpreter'].binary()):
            return '{interpreter} {setup_py} clean {args}'
        return 0  # no need to invoke anything

    @shell_command
    @create_pydistutils_cfg
    def configure(self, context, args):
        return '{interpreter} {setup_py} config {args}'

    @shell_command
    @create_pydistutils_cfg
    def build(self, context, args):
        return '{interpreter.binary_dv} {setup_py} build {args}'

    @shell_command
    def _bdist_wheel(self, context, args):
        # pylint: disable=unused-argument
        try:
            # pylint: disable=unused-import
            import wheel
        except ImportError:
            raise Exception("wheel is required to build wheels for distutils/setuptools packages. Build-Depend on python3-wheel.")
        # Don't build a wheel in the deb install layout
        fpath = join(args['home_dir'], '.pydistutils.cfg')
        remove(fpath)
        return '{interpreter.binary_dv} -c "import setuptools, runpy; runpy.run_path(\'{setup_py}\')" bdist_wheel {args}'

    def build_wheel(self, context, args):
        self._bdist_wheel(context, args)
        dist_dir = join(args['dir'], 'dist')
        wheels = glob1(dist_dir, '*.whl')
        n_wheels = len(wheels)
        if n_wheels != 1:
            raise Exception(f"Expected 1 wheel, found {n_wheels}")
        move(join(dist_dir, wheels[0]), args['home_dir'])

    @shell_command
    @create_pydistutils_cfg
    def install(self, context, args):
        # remove egg-info dirs from build_dir
        for fname in glob1(args['build_dir'], '*.egg-info'):
            fpath = join(args['build_dir'], fname)
            if isdir(fpath):
                rmtree(fpath)
            else:
                remove(fpath)

        return '{interpreter.binary_dv} {setup_py} install --root {destdir} {args}'

    @shell_command
    @create_pydistutils_cfg
    @copy_test_files()
    def test(self, context, args):
        return super().test(context, args)



from tempfile import TemporaryDirectory
import unittest
import os

from dhpython.debhelper import DebHelper, build_options


class DebHelperTestCase(unittest.TestCase):
    impl = 'cpython3'
    control = []
    options = {}
    parse_control = True

    def build_options(self):
        return build_options(**self.options)

    def setUp(self):
        self.tempdir = TemporaryDirectory()  # pylint: disable=consider-using-with
        self.addCleanup(self.tempdir.cleanup)

        old_wd = os.getcwd()
        os.chdir(self.tempdir.name)
        self.addCleanup(os.chdir, old_wd)

        os.mkdir('debian')
        with open('debian/control', 'w', encoding="UTF-8") as f:
            f.write('\n'.join(self.control))
        if self.parse_control:
            self.dh = DebHelper(self.build_options(), impl=self.impl)


CONTROL = [
    'Source: foo-src',
    'Build-Depends: python3-all,',
    ' python-all',
    '\t, bar (<< 2) [amd64],',
    ' baz (>= 1.0)',
    'X-Python3-Version: >= 3.1, << 3.10',
    '',
    'Architecture: all',
    'Package: python3-foo',
    'Depends: ${python3:Depends}',
    '',
    'Package: python3-foo-ext',
    'Architecture: any',
    'Depends: ${python3:Depends}, '
    '# COMMENT',
    ' ${shlibs:Depends},',
    '',
    'Package: python-foo',
    'Architecture: all',
    'Depends: ${python:Depends}',
    '',
    '',
    'Package: foo',
    'Architecture: all',
    'Depends: ${python3:Depends}',
    '',
    '',
    'Package: recfoo',
    'Architecture: all',
    'Recommends: ${python3:Depends}',
    '',
    '',
]

class TestControlBlockParsing(DebHelperTestCase):
    control = CONTROL

    def test_parses_source(self):
        self.assertEqual(self.dh.source_name, 'foo-src')

    def test_parses_build_depends(self):
        self.assertEqual(self.dh.build_depends, {
            'python3-all': {None: None},
            'python-all': {None: None},
            'bar': {'amd64': '<< 2'},
            'baz': {None: '>= 1.0'},
        })

    def test_parses_XPV(self):
        self.assertEqual(self.dh.python_version, '>= 3.1, << 3.10')

    def test_parses_packages(self):
        self.assertEqual(list(self.dh.packages.keys()),
                         ['python3-foo', 'python3-foo-ext', 'foo', 'recfoo'])

    def test_parses_arch(self):
        self.assertEqual(self.dh.packages['python3-foo-ext']['arch'], 'any')

    def test_parses_arch_all(self):
        self.assertEqual(self.dh.packages['python3-foo']['arch'], 'all')


class TestControlSkipIndep(DebHelperTestCase):
    control = CONTROL
    options = {
        'arch': True,
    }

    def test_skip_indep(self):
        self.assertEqual(list(self.dh.packages.keys()), ['python3-foo-ext'])


class TestControlSkipArch(DebHelperTestCase):
    control = CONTROL
    options = {
        'arch': False,
    }

    def test_skip_arch(self):
        self.assertEqual(list(self.dh.packages.keys()),
                         ['python3-foo', 'foo', 'recfoo'])


class TestControlSinglePkg(DebHelperTestCase):
    control = CONTROL
    options = {
        'package': ['python3-foo'],
    }

    def test_parses_packages(self):
        self.assertEqual(list(self.dh.packages.keys()), ['python3-foo'])


class TestControlSkipSinglePkg(DebHelperTestCase):
    control = CONTROL
    options = {
        'no_package': ['python3-foo'],
    }

    def test_parses_packages(self):
        self.assertEqual(list(self.dh.packages.keys()),
                         ['python3-foo-ext', 'foo', 'recfoo'])

class TestControlNoBinaryPackages(DebHelperTestCase):
    control = [
        'Source: foo-src',
        'Build-Depends: python3-all',
        '',
    ]
    parse_control = False

    def test_throws_error(self):
        msg = 'Unable to parse debian/control, found less than 2 paragraphs'
        with self.assertRaisesRegex(Exception, msg):
            DebHelper(self.build_options())


class TestControlMissingPackage(DebHelperTestCase):
    control = [
        'Source: foo-src',
        'Build-Depends: python3-all',
        '',
        'Architecture: all',
    ]
    parse_control = False

    def test_parses_packages(self):
        msg = ('Unable to parse debian/control, paragraph 2 missing Package '
               'field')
        with self.assertRaisesRegex(Exception, msg):
            DebHelper(self.build_options())


class TestRemainingPackages(DebHelperTestCase):
    control = CONTROL
    options = {
        'remaining_packages': True,
    }
    parse_control = False

    def setUp(self):
        super().setUp()
        with open('debian/python3-foo.debhelper.log', 'w', encoding="UTF-8") as f:
            f.write('dh_python3\n')
        with open('debian/python3-foo-ext.debhelper.log', 'w', encoding="UTF-") as f:
            f.write('dh_foobar\n')
        self.dh = DebHelper(self.build_options(), impl=self.impl)

    def test_skips_logged_packages(self):
        self.assertEqual(list(self.dh.packages.keys()),
                         ['python3-foo-ext', 'foo', 'recfoo'])



from setuptools import setup, Extension

setup_args = dict(
    ext_modules = [
        Extension(
            'foo.ext',
            ['foo/ext.c'],
            py_limited_api = True,
        )
    ]
)
setup(**setup_args)



# Copyright © 2012-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from dhpython.build.base import Base, shell_command, copy_test_files


class BuildSystem(Base):
    DESCRIPTION = 'use --*-args options to configure this system'
    SUPPORTED_INTERPRETERS = True  # all interpreters

    @shell_command
    def clean(self, context, args):
        super().clean(context, args)
        return args['args']

    @shell_command
    def configure(self, context, args):
        return args['args']

    @shell_command
    def build(self, context, args):
        return args['args']

    @shell_command
    def install(self, context, args):
        return args['args']

    @shell_command
    @copy_test_files()
    def test(self, context, args):
        return args['args'] or super().test(context, args)



import os
import logging
import platform
import unittest
from copy import deepcopy
from pickle import dumps
from tempfile import TemporaryDirectory

from dhpython.depends import Dependencies

from .common import FakeOptions


def pep386(d):
    """Mark all pydist entries as being PEP386"""
    for k, v in d.items():
        if isinstance(v, str):
            d[k] = {'dependency': v}
            d[k].setdefault('standard', 'PEP386')
    return d


def prime_pydist(impl, pydist):
    """Fake the pydist data for impl. Returns a cleanup function"""
    from dhpython.pydist import load

    for name, entries in pydist.items():
        if not isinstance(entries, list):
            pydist[name] = entries = [entries]
        for i, entry in enumerate(entries):
            if isinstance(entry, str):
                entries[i] = entry = {'dependency': entry}
            entry.setdefault('name', name)
            entry.setdefault('standard', '')
            entry.setdefault('rules', [])
            entry.setdefault('versions', set())

    key = dumps(((impl,), {}))
    load.cache[key] = pydist
    return lambda: load.cache.pop(key)


class DependenciesTestCase(unittest.TestCase):
    pkg = 'foo'
    impl = 'cpython3'
    pydist = {}
    stats = {
        'compile': False,
        'dist-info': set(),
        'egg-info': set(),
        'ext_no_version': set(),
        'ext_vers': set(),
        'nsp.txt': set(),
        'private_dirs': {},
        'public_vers': set(),
        'requires.txt': set(),
        'shebangs': set(),
    }
    requires = {}
    dist_info_metadata = {}
    options = FakeOptions()
    parse = True

    def setUp(self):
        self.d = Dependencies(self.pkg, self.impl)

        stats = deepcopy(self.stats)
        write_files = {}
        if self.requires:
            for fn, lines in self.requires.items():
                write_files[fn] = lines
                stats['requires.txt'].add(fn)

        if self.dist_info_metadata:
            for fn, lines in self.dist_info_metadata.items():
                write_files[fn] = lines
                stats['dist-info'].add(fn)

        if write_files:
            self.tempdir = TemporaryDirectory()  # pylint: disable=consider-using-with
            self.addCleanup(self.tempdir.cleanup)
            old_wd = os.getcwd()
            os.chdir(self.tempdir.name)
            self.addCleanup(os.chdir, old_wd)

        for fn, lines in write_files.items():
            os.makedirs(os.path.dirname(fn))
            with open(fn, 'w', encoding="UTF-8") as f:
                f.write('\n'.join(lines))

        cleanup = prime_pydist(self.impl, self.pydist)
        self.addCleanup(cleanup)

        if self.parse:
            self.d.parse(stats, self.options)
        else:
            self.prepared_stats = stats

    def assertNotInDepends(self, pkg):
        """Assert that pkg doesn't appear *anywhere* in self.d.depends"""
        for dep in self.d.depends:
            for alt in dep.split('|'):
                alt = alt.strip().split('(', 1)[0].strip()
                if pkg == alt:
                    raise AssertionError(f'{pkg} appears in {alt}')


class TestRequiresCPython3(DependenciesTestCase):
    options = FakeOptions(guess_deps=True)
    pydist = {
        'bar': 'python3-bar',
        'baz': {'dependency': 'python3-baz', 'standard': 'PEP386'},
        'quux': {'dependency': 'python3-quux', 'standard': 'PEP440'},
    }
    requires = {
        'debian/foo/usr/lib/python3/dist-packages/foo.egg-info/requires.txt': (
            'bar',
            'baz >= 1.0',
            'quux >= 1.0a1',
        ),
    }

    def test_depends_on_bar(self):
        self.assertIn('python3-bar', self.d.depends)

    def test_depends_on_baz(self):
        self.assertIn('python3-baz (>= 1.0)', self.d.depends)

    def test_depends_on_quux(self):
        self.assertIn('python3-quux (>= 1.0~a1)', self.d.depends)


class TestRequiresCompatible(DependenciesTestCase):
    options = FakeOptions(guess_deps=True)
    pydist = {
        'bar': 'python3-bar',
        'baz': {'dependency': 'python3-baz', 'standard': 'PEP386'},
        'qux': {'dependency': 'python3-qux', 'standard': 'PEP386'},
        'quux': {'dependency': 'python3-quux', 'standard': 'PEP386'},
        'corge': {'dependency': 'python3-corge', 'standard': 'PEP440'},
    }
    requires = {
        'debian/foo/usr/lib/python3/dist-packages/foo.egg-info/requires.txt': (
            'bar',
            'baz ~= 1.1',
            'qux == 1.*',
            'quux',
            'corge == 1.0',
        ),
    }

    def test_depends_on_bar(self):
        self.assertIn('python3-bar', self.d.depends)

    def test_depends_on_baz(self):
        self.assertIn('python3-baz (>= 1.1), python3-baz (<< 2)', self.d.depends)

    def test_depends_on_qux(self):
        self.assertIn('python3-qux (>= 1.0), python3-qux (<< 2)', self.d.depends)

    def test_depends_on_corge(self):
        self.assertIn('python3-corge (>= 1.0), python3-corge (<< 1.1~)',
                      self.d.depends)


class TestRequiresDistPython3(DependenciesTestCase):
    options = FakeOptions(guess_deps=True)
    pydist = {
        'bar': 'python3-bar',
        'baz': {'dependency': 'python3-baz', 'standard': 'PEP386'},
        'qux': {'dependency': 'python3-qux', 'standard': 'PEP386'},
        'quux': {'dependency': 'python3-quux', 'standard': 'PEP386'},
        'corge': {'dependency': 'python3-corge', 'standard': 'PEP440'},
    }
    dist_info_metadata = {
        'debian/foo/usr/lib/python3/dist-packages/foo.dist-info/METADATA': (
            'Requires-Dist: bar',
            'Requires-Dist: baz >= 1.0',
            'Requires-Dist: qux == 1.*',
            'Requires-Dist: quux ~= 1.1',
            'Requires-Dist: corge == 1.0',
        ),
    }

    def test_depends_on_bar(self):
        self.assertIn('python3-bar', self.d.depends)

    def test_depends_on_baz(self):
        self.assertIn('python3-baz (>= 1.0)', self.d.depends)

    def test_depends_on_qux(self):
        self.assertIn('python3-qux (>= 1.0), python3-qux (<< 2)',
                      self.d.depends)

    def test_depends_on_quux(self):
        self.assertIn('python3-quux (>= 1.1), python3-quux (<< 2)',
                      self.d.depends)

    def test_depends_on_corge(self):
        self.assertIn('python3-corge (>= 1.0), python3-corge (<< 1.1~)',
                      self.d.depends)


class TestEnvironmentMarkersDistInfo(DependenciesTestCase):
    options = FakeOptions(guess_deps=True, depends_section=['feature'])
    pydist = pep386({
        'no_markers': 'python3-no-markers',
        'os_posix': 'python3-os-posix',
        'os_java': 'python3-os-java',
        'sys_platform_linux': 'python3-sys-platform-linux',
        'sys_platform_darwin': 'python3-sys-platform-darwin',
        'platform_machine_x86_64': 'python3-platform-machine-x86-64',
        'platform_machine_mips64': 'python3-platform-machine-mips64',
        'platform_python_implementation_cpython':
            'python3-platform-python-implementation-cpython',
        'platform_python_implementation_jython':
            'python3-platform-python-implementation-jython',
        'platform_release_lt2': 'python3-platform-release-lt2',
        'platform_release_ge2': 'python3-platform-release-ge2',
        'platform_system_linux': 'python3-platform-system-linux',
        'platform_system_windows': 'python3-platform-system-windows',
        'platform_version_lt1': 'python3-platform-version-lt1',
        'platform_version_ge1': 'python3-platform-version-ge1',
        'python_version_ge3': 'python3-python-version-ge3',
        'python_version_gt3': 'python3-python-version-gt3',
        'python_version_lt3': 'python3-python-version-lt3',
        'python_version_lt30': 'python3-python-version-lt30',
        'python_version_lt38': 'python3-python-version-lt38',
        'python_version_lt313': 'python3-python-version-lt313',
        'python_version_le313': 'python3-python-version-le313',
        'python_version_ge27': 'python3-python-version-ge27',
        'python_version_ge313': 'python3-python-version-ge313',
        'python_version_gt313': 'python3-python-version-gt313',
        'python_version_eq313': 'python3-python-version-eq313',
        'python_version_ne313': 'python3-python-version-ne313',
        'python_version_aeq313': 'python3-python-version-aeq313',
        'python_version_ceq313': 'python3-python-version-ceq313',
        'python_version_weq313': 'python3-python-version-weq313',
        'python_version_full_lt300': 'python3-python-version-full-lt300',
        'python_version_full_lt3131': 'python3-python-version-full-lt3131',
        'python_version_full_le3131': 'python3-python-version-full-le3131',
        'python_version_full_ge3131': 'python3-python-version-full-ge3131',
        'python_version_full_ge3131a1': 'python3-python-version-full-ge3131a1',
        'python_version_full_ge3131b1post1':
            'python3-python-version-full-ge3131b1post1',
        'python_version_full_gt3131': 'python3-python-version-full-gt3131',
        'python_version_full_eq3131': 'python3-python-version-full-eq3131',
        'python_version_full_ne3131': 'python3-python-version-full-ne3131',
        'python_version_full_aeq3131': 'python3-python-version-full-aeq3131',
        'python_version_full_ceq3131': 'python3-python-version-full-ceq3131',
        'python_version_full_weq313': 'python3-python-version-full-weq313',
        'implementation_name_cpython': 'python3-implementation-name-cpython',
        'implementation_name_pypy': 'python3-implementation-name-pypy',
        'implementation_version_lt313': 'python3-implementation-version-lt313',
        'implementation_version_ge313': 'python3-implementation-version-ge313',
        'invalid_marker': 'python3-invalid-marker',
        'extra_feature': 'python3-extra-feature',
        'extra_test': 'python3-extra-test',
        'complex_marker': 'python3-complex-marker',
        'complex_marker_2': 'python3-complex-marker-2',
        'no_markers_2': 'python3-no-markers-2',
    })
    dist_info_metadata = {
        'debian/foo/usr/lib/python3/dist-packages/foo.dist-info/METADATA': (
            "Requires-Dist: no_markers",
            "Requires-Dist: os_posix; (os_name == 'posix')",
            'Requires-Dist: os_java; os_name == "java"',
            "Requires-Dist: sys_platform_linux ; sys_platform == 'linux'",
            "Requires-Dist: sys_platform_darwin;sys_platform == 'darwin'",
            "Requires-Dist: platform_machine_x86_64; "
                "platform_machine == 'x86_64'",
            "Requires-Dist: platform_machine_mips64; "
                "platform_machine == 'mips64'",
            "Requires-Dist: platform_python_implementation_cpython; "
                "platform_python_implementation == 'CPython'",
            "Requires-Dist: platform_python_implementation_jython; "
                "platform_python_implementation == 'Jython'",
            "Requires-Dist: platform_release_lt2; platform_release < '2.0'",
            "Requires-Dist: platform_release_ge2; platform_release >= '2.0'",
            "Requires-Dist: platform_system_linux; platform_system == 'Linux'",
            "Requires-Dist: platform_system_windows; "
                "platform_system == 'Windows'",
            "Requires-Dist: platform_version_lt1; platform_version < '1'",
            "Requires-Dist: platform_version_ge1; platform_version >= '1'",
            "Requires-Dist: python_version_ge3; python_version >= '3'",
            "Requires-Dist: python_version_gt3; python_version > '3'",
            "Requires-Dist: python_version_lt3; python_version < '3'",
            "Requires-Dist: python_version_lt30; python_version < '3.0'",
            "Requires-Dist: python_version_lt38; python_version < '3.8'",
            "Requires-Dist: python_version_lt313; python_version < '3.13'",
            "Requires-Dist: python_version_le313; python_version <= '3.13'",
            "Requires-Dist: python_version_gt313; python_version > '3.13'",
            "Requires-Dist: python_version_ge27; python_version >= '2.7'",
            "Requires-Dist: python_version_ge313; python_version >= '3.13'",
            "Requires-Dist: python_version_eq313; python_version == '3.13'",
            "Requires-Dist: python_version_ne313; python_version != '3.13'",
            "Requires-Dist: python_version_aeq313; python_version === '3.13'",
            "Requires-Dist: python_version_ceq313; python_version ~= '3.13'",
            "Requires-Dist: python_version_weq313; python_version == '3.13.*'",
            "Requires-Dist: python_version_full_lt300; "
                "python_full_version < '3.0.0'",
            "Requires-Dist: python_version_full_lt3131; "
                "python_full_version < '3.13.1'",
            "Requires-Dist: python_version_full_le3131; "
                "python_full_version <= '3.13.1'",
            "Requires-Dist: python_version_full_gt3131; "
                "python_full_version > '3.13.1'",
            "Requires-Dist: python_version_full_ge3131; "
                "python_full_version >= '3.13.1'",
            "Requires-Dist: python_version_full_ge3131a1; "
                "python_full_version >= '3.13.1a1'",
            "Requires-Dist: python_version_full_ge3131b1post1; "
                "python_full_version >= '3.13.1b1.post1'",
            "Requires-Dist: python_version_full_eq3131; "
                "python_full_version == '3.13.1'",
            "Requires-Dist: python_version_full_ne3131; "
                "python_full_version != '3.13.1'",
            "Requires-Dist: python_version_full_aeq3131; "
                "python_full_version === '3.13.1'",
            "Requires-Dist: python_version_full_ceq3131; "
                "python_full_version ~= '3.13.1'",
            "Requires-Dist: python_version_full_weq313; "
                "python_full_version == '3.13.*'",
            "Requires-Dist: implementation_name_cpython; "
                "implementation_name == 'cpython'",
            "Requires-Dist: implementation_name_pypy; "
                "implementation_name == 'pypy'",
            "Requires-Dist: implementation_version_lt313; "
                "implementation_version < '3.13'",
            "Requires-Dist: implementation_version_ge313; "
                "implementation_version >= '3.13'",
            "Requires-Dist: invalid_marker; invalid_marker > '1'",
            "Requires-Dist: extra_feature; extra == 'feature'",
            "Requires-Dist: extra_test; extra == 'test'",
            "Requires-Dist: complex_marker; os_name != 'windows' "
                "and implementation_name == 'cpython'",
            "Requires-Dist: complex_marker_2; (python_version > \"3.4\") "
                "and extra == 'test'",
            "Requires-Dist: no_markers_2",
        ),
    }

    def test_depends_on_unmarked_packages(self):
        self.assertIn('python3-no-markers', self.d.depends)

    def test_depends_on_posix_packages(self):
        self.assertIn('python3-os-posix', self.d.depends)

    def test_skips_non_posix_packages(self):
        self.assertNotInDepends('python3-os-java')

    def test_depends_on_linux_packages(self):
        self.assertIn('python3-sys-platform-linux', self.d.depends)

    def test_skips_darwin_packages(self):
        self.assertNotInDepends('python3-sys-platform-darwin')

    def test_depends_on_x86_64_packages_on_x86_64(self):
        if platform.machine() == 'x86_64':
            self.assertIn('python3-platform-machine-x86-64', self.d.depends)
        else:
            self.assertNotInDepends('python3-platform-machine-x86-64')

    def test_depends_on_mips64_packages_on_mips64(self):
        if platform.machine() == 'mips64':
            self.assertIn('python3-platform-machine-mips64', self.d.depends)
        else:
            self.assertNotInDepends('python3-platform-machine-mips64')

    def test_depends_on_plat_cpython_packages(self):
        self.assertIn('python3-platform-python-implementation-cpython',
                      self.d.depends)

    def test_skips_plat_jython_packages(self):
        self.assertNotInDepends('python3-platform-python-implementation-jython')

    def test_skips_release_lt_2_packages(self):
        self.assertNotInDepends('python3-platform-release-lt2')

    def test_skips_release_gt_2_packages(self):
        self.assertNotInDepends('python3-platform-release-ge2')

    def test_depends_on_platform_linux_packages(self):
        self.assertIn('python3-platform-system-linux', self.d.depends)

    def test_skips_platform_windows_packages(self):
        self.assertNotInDepends('python3-platform-system-windows')

    def test_skips_platfrom_version_lt_1_packages(self):
        self.assertNotInDepends('python3-platform-version-lt1')

    def test_skips_platform_version_ge_1_packages(self):
        self.assertNotInDepends('python3-platform-version-ge1')

    def test_skips_py_version_lt_3_packages(self):
        self.assertNotInDepends('python3-python-version-lt3')

    def test_elides_py_version_ge_3(self):
        self.assertIn('python3-python-version-ge3', self.d.depends)

    def test_elides_py_version_gt_3(self):
        self.assertIn('python3-python-version-gt3', self.d.depends)

    def test_skips_py_version_lt_30_packages(self):
        self.assertNotInDepends('python3-python-version-lt30')

    def test_skips_py_version_lt_38_packages(self):
        self.assertNotInDepends('python3-python-version-lt38')

    def test_depends_on_py_version_lt_313_packages(self):
        self.assertIn('python3-python-version-lt313 '
                      '| python3-supported-min (>= 3.13)', self.d.depends)

    def test_depends_on_py_version_le_313_packages(self):
        self.assertIn('python3-python-version-le313 '
                      '| python3-supported-min (>> 3.13)', self.d.depends)

    def test_depends_on_py_version_ge_27_packages(self):
        self.assertIn('python3-python-version-ge27',
                      self.d.depends)

    def test_depends_on_py_version_ge_313_packages(self):
        self.assertIn('python3-python-version-ge313 '
                      '| python3-supported-max (<< 3.13)', self.d.depends)

    def test_depends_on_py_version_gt_313_packages(self):
        self.assertIn('python3-python-version-gt313 '
                      '| python3-supported-max (<= 3.13)', self.d.depends)

    def test_depends_on_py_version_eq_313_packages(self):
        self.assertIn(
            'python3-python-version-eq313 | python3-supported-max (<< 3.13) '
            '| python3-supported-min (>= 3.14)', self.d.depends)

    def test_depends_on_py_version_ne_313_packages(self):
        # Can't be represented in Debian depends
        self.assertIn('python3-python-version-ne313', self.d.depends)

    def test_depends_on_py_version_aeq_313_packages(self):
        self.assertIn(
            'python3-python-version-aeq313 | python3-supported-max (<< 3.13) '
            '| python3-supported-min (>> 3.13)', self.d.depends)

    def test_depends_on_py_version_ceq_313_packages(self):
        self.assertIn(
            'python3-python-version-ceq313 | python3-supported-max (<< 3.13) '
            '| python3-supported-min (>= 3.14)', self.d.depends)

    def test_depends_on_py_version_weq_313_packages(self):
        self.assertIn(
            'python3-python-version-weq313 | python3-supported-max (<< 3.13) '
            '| python3-supported-min (>= 3.14)', self.d.depends)

    def test_skips_py_version_full_lt_300_packages(self):
        self.assertNotInDepends('python3-python-version-full-lt300')

    def test_depends_on_py_version_full_lt_3131_packages(self):
        self.assertIn('python3-python-version-full-lt3131 '
                      '| python3-supported-min (>= 3.13.1)', self.d.depends)

    def test_depends_on_py_version_full_le_3131_packages(self):
        self.assertIn('python3-python-version-full-le3131 '
                      '| python3-supported-min (>> 3.13.1)', self.d.depends)

    def test_depends_on_py_version_full_ge_3131_packages(self):
        self.assertIn('python3-python-version-full-ge3131 '
                      '| python3-supported-max (<< 3.13.1)', self.d.depends)

    def test_depends_on_py_version_full_ge_3131a1_packages(self):
        # With full PEP-440 parsing this should be (<< 3.13.1~a1)
        self.assertIn('python3-python-version-full-ge3131a1 '
                      '| python3-supported-max (<< 3.13.0)', self.d.depends)

    def test_depends_on_py_version_full_ge_3131b1post1_packages(self):
        # With full PEP-440 parsing this should be (<< 3.13.1~b1.post1)
        self.assertIn('python3-python-version-full-ge3131a1 '
                      '| python3-supported-max (<< 3.13.0)', self.d.depends)

    def test_depends_on_py_version_full_gt_3131_packages(self):
        self.assertIn('python3-python-version-full-gt3131 '
                      '| python3-supported-max (<= 3.13.1)', self.d.depends)

    def test_depends_on_py_version_full_eq_3131_packages(self):
        self.assertIn('python3-python-version-full-eq3131 '
                      '| python3-supported-max (<< 3.13.1) '
                      '| python3-supported-min (>> 3.13.1)', self.d.depends)

    def test_depends_on_py_version_full_ne_3131_packages(self):
        # Can't be represented in Debian depends
        self.assertIn('python3-python-version-full-ne3131', self.d.depends)

    def test_skips_py_version_full_aeq_3131_packages(self):
        # Can't be represented in Debian depends
        self.assertNotInDepends('python3-python-version-full-aeq3131')

    def test_depends_on_py_version_full_ceq_3131_packages(self):
        self.assertIn('python3-python-version-full-ceq3131 '
                      '| python3-supported-max (<< 3.13.1) '
                      '| python3-supported-min (>= 3.14)', self.d.depends)

    def test_depends_on_py_version_full_weq_313_packages(self):
        self.assertIn('python3-python-version-full-weq313 '
                      '| python3-supported-max (<< 3.13) '
                      '| python3-supported-min (>= 3.14)', self.d.depends)

    def test_depends_on_sys_cpython_packages(self):
        self.assertIn('python3-implementation-name-cpython', self.d.depends)

    def test_depends_on_sys_pypy_packages(self):
        self.assertIn('python3-implementation-name-pypy', self.d.depends)

    def test_depends_on_sys_implementation_lt313_packages(self):
        self.assertIn('python3-implementation-version-lt313 '
                      '| python3-supported-min (>= 3.13)',
                      self.d.depends)

    def test_depends_on_sys_implementation_ge313_packages(self):
        self.assertIn('python3-implementation-version-ge313 '
                      '| python3-supported-max (<< 3.13)',
                      self.d.depends)

    def test_ignores_invalid_marker(self):
        self.assertNotInDepends('python3-invalid-marker')

    def test_depends_on_extra_feature_packages(self):
        self.assertIn('python3-extra-feature', self.d.depends)

    def test_skips_extra_test_packages(self):
        self.assertNotInDepends('python3-extra-test')

    def test_ignores_complex_environment_markers(self):
        self.assertNotInDepends('python3-complex-marker')
        self.assertNotInDepends('python3-complex-marker-2')

    def test_depends_on_un_marked_dependency_after_extra(self):
        self.assertIn('python3-no-markers-2', self.d.depends)


class TestEnvironmentMarkersEggInfo(TestEnvironmentMarkersDistInfo):
    dist_info_metadata = None
    requires = {
        'debian/foo/usr/lib/python3/dist-packages/foo.egg-info/requires.txt': (
            "no_markers",
            "[:(os_name == 'posix')]",
            "os_posix",
            '[:os_name == "java"]',
            "os_java",
            "[:sys_platform == 'linux']",
            "sys_platform_linux",
            "[:sys_platform == 'darwin']",
            "sys_platform_darwin",
            "[:platform_machine == 'x86_64']",
            "platform_machine_x86_64",
            "[:platform_machine == 'mips64']",
            "platform_machine_mips64",
            "[:platform_python_implementation == 'CPython']",
            "platform_python_implementation_cpython",
            "[:platform_python_implementation == 'Jython']",
            "platform_python_implementation_jython",
            "[:platform_release < '2.0']",
            "platform_release_lt2",
            "[:platform_release >= '2.0']",
            "platform_release_ge2",
            "[:platform_system == 'Linux']",
            "platform_system_linux",
            "[:platform_system == 'Windows']",
            "platform_system_windows",
            "[:platform_version < '1']",
            "platform_version_lt1",
            "[:platform_version >= '1']",
            "platform_version_ge1",
            "[:python_version >= '3']",
            "python_version_ge3",
            "[:python_version > '3']",
            "python_version_gt3",
            "[:python_version < '3']",
            "python_version_lt3",
            "[:python_version < '3.0']",
            "python_version_lt30",
            "[:python_version < '3.8']",
            "python_version_lt313",
            "[:python_version < '3.13']",
            "python_version_lt313",
            "[:python_version <= '3.13']",
            "python_version_le313",
            "[:python_version > '3.13']",
            "python_version_gt313",
            "[:python_version >= '2.7']",
            "python_version_ge27",
            "[:python_version >= '3.13']",
            "python_version_ge313",
            "[:python_version == '3.13']",
            "python_version_eq313",
            "[:python_version != '3.13']",
            "python_version_ne313",
            "[:python_version === '3.13']",
            "python_version_aeq313",
            "[:python_version ~= '3.13']",
            "python_version_ceq313",
            "[:python_version == '3.13.*']",
            "python_version_weq313",
            "[:python_full_version < '3.0.0']",
            "python_version_full_lt300",
            "[:python_full_version < '3.13.1']",
            "python_version_full_lt3131",
            "[:python_full_version <= '3.13.1']",
            "python_version_full_le3131",
            "[:python_full_version > '3.13.1']",
            "python_version_full_gt3131",
            "[:python_full_version >= '3.13.1']",
            "python_version_full_ge3131",
            "[:python_full_version >= '3.13.1a1']",
            "python_version_full_ge3131a1",
            "[:python_full_version >= '3.13.1b1.post1']",
            "python_version_full_ge3131b1post1",
            "[:python_full_version == '3.13.1']",
            "python_version_full_eq3131",
            "[:python_full_version != '3.13.1']",
            "python_version_full_ne3131",
            "[:python_full_version === '3.13.1']",
            "python_version_full_aeq3131",
            "[:python_full_version ~= '3.13.1']",
            "python_version_full_ceq3131",
            "[:python_full_version == '3.13.*']",
            "python_version_full_weq313",
            "[:implementation_name == 'cpython']",
            "implementation_name_cpython",
            "[:implementation_name == 'pypy']",
            "implementation_name_pypy",
            "[:implementation_version < '3.13']",
            "implementation_version_lt313",
            "[:implementation_version >= '3.13']",
            "implementation_version_ge313",
            "[:invalid_marker > '1']",
            "invalid_marker",
            "[feature]",
            "extra_feature",
            "[test]",
            "extra_test",
            "[:os_name != 'windows' and implementation_name == 'cpython']",
            "complex_marker",
            "[test:(os_name != 'windows')]",
            "complex_marker_2",
        ),
    }

    def test_depends_on_un_marked_dependency_after_extra(self):
        raise unittest.SkipTest('Not possible in requires.txt')


class TestIgnoresUnusedModulesDistInfo(DependenciesTestCase):
    options = FakeOptions(guess_deps=True, depends_section=['feature'])
    dist_info_metadata = {
        'debian/foo/usr/lib/python3/dist-packages/foo.dist-info/METADATA': (
            "Requires-Dist: unusued-complex-module ; "
                "(sys_platform == \"darwin\") and extra == 'nativelib'",
            "Requires-Dist: unused-win-module ; (sys_platform == \"win32\")",
            "Requires-Dist: unused-extra-module ; extra == 'unused'",
        ),
    }
    parse = False

    def test_ignores_unused_dependencies(self):
        if not hasattr(self, 'assertLogs'):
            raise unittest.SkipTest("Requires Python >= 3.4")
        with self.assertLogs(logger='dhpython', level=logging.INFO) as logs:
            self.d.parse(self.prepared_stats, self.options)
        for line in logs.output:
            self.assertTrue(
                line.startswith(
                    'INFO:dhpython:Ignoring complex environment marker'),
                'Expecting only complex environment marker messages, but '
                'got: {}'.format(line))


class TestIgnoresUnusedModulesEggInfo(DependenciesTestCase):
    options = FakeOptions(guess_deps=True, depends_section=['feature'])
    requires = {
        'debian/foo/usr/lib/python3/dist-packages/foo.egg-info/requires.txt': (
            "[nativelib:(sys_platform == 'darwin')]",
            "unusued-complex-module",
            "[:sys_platform == 'win32']",
            "unused-win-module",
            "[unused]",
            "unused-extra-module",
        )
    }
    parse = False

    def test_ignores_unused_dependencies(self):
        if not hasattr(self, 'assertNoLogs'):
            raise unittest.SkipTest("Requires Python >= 3.10")
        with self.assertNoLogs(logger='dhpython', level=logging.INFO):
            self.d.parse(self.prepared_stats, self.options)



class FakeOptions:
    def __init__(self, **kwargs):
        opts = {
            'depends': (),
            'depends_section': (),
            'guess_deps': False,
            'no_ext_rename': False,
            'recommends': (),
            'recommends_section': (),
            'requires': (),
            'suggests': (),
            'suggests_section': (),
            'vrange': None,
            'accept_upstream_versions': False,
        }
        opts.update(kwargs)
        for k, v in opts.items():
            setattr(self, k, v)



# Copyright © 2023 Stefano Rivera <stefanor@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os.path import join

from dhpython.build.base import Base, shell_command, copy_test_files


class BuildSystem(Base):
    DESCRIPTION = 'meson build system (using dh_auto_* commands)'
    REQUIRED_COMMANDS = ['meson']
    REQUIRED_FILES = ['meson.build']

    @shell_command
    def clean(self, context, args):
        super().clean(context, args)
        return 'dh_auto_clean --buildsystem=cmake'

    @shell_command
    def configure(self, context, args):
        # Can't be specified on the command line, directly
        # https://github.com/mesonbuild/meson/issues/9671
        with open(join(args['build_dir'], 'pybuild-meson-native.ini'), 'w', encoding="UTF-8") as f:
            f.write('[binaries]\n')
            f.write("python3 = '" + args['interpreter'].binary_dv + "'\n")

        return ('dh_auto_configure --buildsystem=meson'
                ' --builddirectory={build_dir} --'
                ' --native-file={build_dir}/pybuild-meson-native.ini'
                ' {args}')

    @shell_command
    def build(self, context, args):
        return ('dh_auto_build --buildsystem=meson'
                ' --builddirectory={build_dir}'
                ' -- {args}')

    @shell_command
    def install(self, context, args):
        return ('dh_auto_install --buildsystem=meson'
                ' --builddirectory={build_dir}'
                ' --destdir={destdir}'
                ' -- {args}')

    @shell_command
    @copy_test_files()
    def test(self, context, args):
        return ('dh_auto_test --buildsystem=meson'
                ' --builddirectory={build_dir}'
                ' -- {args}')



from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import TestCase

from dhpython.interpreter import Interpreter
from dhpython.fs import merge_WHEEL, share_files

from .common import FakeOptions


class MergeWheelTestCase(TestCase):
    files = {}
    def setUp(self):
        self.tempdir = TemporaryDirectory()  # pylint: disable=consider-using-with
        self.addCleanup(self.tempdir.cleanup)
        temp_path = Path(self.tempdir.name)
        for fn, contents in self.files.items():
            path = temp_path / fn
            setattr(self, path.name, path)
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open('w') as f:
                f.write('\n'.join(contents))
                f.write('\n')

    def assertFileContents(self, path, contents):
        """Assert that the contents of path is contents

        Contents may be specified as a list of strings, one per line, without
        line-breaks.
        """
        if isinstance(contents, (list, tuple)):
            contents = '\n'.join(contents) + '\n'
        with path.open('r') as f:
            self.assertMultiLineEqual(contents, f.read())


class MergeTagsTest(MergeWheelTestCase):
    files = {
        'a': ('foo', 'Tag: A'),
        'b': ('foo', 'Tag: B'),
    }

    def test_merge_wheel(self):
        merge_WHEEL(self.a, self.b)
        self.assertFileContents(self.b, ('foo', 'Tag: B', 'Tag: A'))


class ShareFilesTestCase(MergeWheelTestCase):
    impl = 'cpython3'
    options = {}

    def setUp(self):
        super().setUp()
        self.destdir = TemporaryDirectory()  # pylint: disable=consider-using-with
        self.addCleanup(self.destdir.cleanup)
        self.interpreter = Interpreter(self.impl)

    def share_files(self, options=None):
        if options is None:
            options = self.options
        share_files(self.tempdir.name, self.destdir.name, self.interpreter,
                    FakeOptions(**options))

    def destPath(self, name):
        return Path(self.destdir.name) / name


class HatchlingLicenseTest(ShareFilesTestCase):
    files = {
        'foo.dist-info/license_files/LICENSE.txt': ('foo'),
        'foo.dist-info/licenses/COPYING': ('foo'),
        'foo.dist-info/RECORD': (
            'foo.dist-info/license_files/LICENSE.txt,sha256=2c26b46b68ffc68ff99'
            'b453c1d30413413422d706483bfa0f98a5e886266e7ae,4',
            'foo.dist-info/licenses/COPYING,sha256=2c26b46b68ffc68ff99b453c1d30'
            '413413422d706483bfa0f98a5e886266e7ae,4',
            'foo.dist-info/WHEEL,sha256=447fb61fa39a067229e1cce8fc0953bfced53ea'
            'c85d1844f5940f51c1fcba725,6'),
        'foo.dist-info/WHEEL': ('foo'),
    }

    def test_removes_license_files(self):
        self.share_files()
        self.assertFalse(
            self.destPath('foo.dist-info/license_files/LICENSE.txt').exists())
        self.assertFalse(
            self.destPath('foo.dist-info/licenses/COPYING').exists())


class FlitLicenseTest(ShareFilesTestCase):
    files = {
        'foo.dist-info/COPYING': ('foo'),
        'foo.dist-info/COPYING.LESSER': ('foo'),
        'foo.dist-info/RECORD': (
            'foo.dist-info/COPYING,sha256=2c26b46b68ffc68ff99b453c1d30413413422'
            'd706483bfa0f98a5e886266e7ae,4',
            'foo.dist-info/COPYING.LESSER,sha256=2c26b46b68ffc68ff99b453c1d3041'
            '3413422d706483bfa0f98a5e886266e7ae,4',
            'foo.dist-info/WHEEL,sha256=447fb61fa39a067229e1cce8fc0953bfced53ea'
            'c85d1844f5940f51c1fcba725,6'),
        'foo.dist-info/WHEEL': ('foo'),
    }

    def test_removes_license_files(self):
        self.share_files()
        self.assertFalse(self.destPath('foo.dist-info/COPYING.LESSER').exists())
        self.assertFalse(self.destPath('foo.dist-info/COPYING').exists())


class CExtensionsRenameTest(ShareFilesTestCase):
    def setUp(self):
        ver = Interpreter(self.impl).default_version
        self.files = {
            f'usr/lib/python{ver}/foo.so': ('binary-data'),
        }
        self.ext_filename = f'usr/lib/python{ver}/foo.so'
        super().setUp()

    def test_python_extensions(self):
        self.share_files()
        self.assertFalse(self.destPath(self.ext_filename).exists())
        expected_name = self.interpreter.check_extname(self.ext_filename)
        self.assertTrue(self.destPath(expected_name).exists())

    def test_no_ext_rename(self):
        self.share_files(options={"no_ext_rename": True})
        self.assertTrue(self.destPath(self.ext_filename).exists())



#! /usr/bin/python3
from distutils.core import setup

setup(name='Foo',
      version='0.2',
      description="package with public modules only",
      long_description="TODO",
      keywords='foo bar baz',
      author='Piotr Ożarowski',
      author_email='piotr@debian.org',
      url='http://www.debian.org/',
      license='MIT',
      packages=['foo'],
      package_dir={'foo': 'lib/foo'},
      package_data={'foo': ['jquery.js']},
      zip_safe=False,
)



# -*- coding: UTF-8 -*-
# Copyright © 2010-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
import os
import re
from datetime import datetime
from glob import glob
from pickle import dumps
from shutil import rmtree
from os.path import exists, getsize, isdir, islink, join, split
from subprocess import Popen, PIPE

log = logging.getLogger('dhpython')
EGGnPTH_RE = re.compile(r'(.*?)(-py\d\.\d(?:-[^.]*)?)?(\.egg-info|\.pth)$')
SHAREDLIB_RE = re.compile(r'NEEDED.*libpython(\d\.\d)')


def relpath(target, link):
    """Return relative path.

    >>> relpath('/usr/share/python-foo/foo.py', '/usr/bin/foo', )
    '../share/python-foo/foo.py'
    """
    t = target.split('/')
    l = link.split('/')
    while l and l[0] == t[0]:
        del l[0], t[0]
    return '/'.join(['..'] * (len(l) - 1) + t)


def relative_symlink(target, link):
    """Create relative symlink."""
    return os.symlink(relpath(target, link), link)


def move_file(fpath, dstdir):
    """Move file to dstdir. Works with symlinks (including relative ones)."""
    if isdir(fpath):
        dname = split(fpath)[-1]
        for fn in os.listdir(fpath):
            move_file(join(fpath, fn), join(dstdir, dname))

    if islink(fpath):
        dstpath = join(dstdir, split(fpath)[-1])
        relative_symlink(os.readlink(fpath), dstpath)
        os.remove(fpath)
    else:
        os.rename(fpath, dstdir)


def move_matching_files(src, dst, pattern, sub=None, repl=''):
    r"""Move files (preserving path) that match given pattern.

    move_matching_files('foo/bar/', 'foo/baz/', r'spam/.*\.so$')
    will move foo/bar/a/b/c/spam/file.so to foo/baz/a/b/c/spam/file.so

    :param sub: regular expression for path part that will be replaced with `repl`
    :param repl: replacement for `sub`
    """
    match = re.compile(pattern).search
    if sub:
        sub = re.compile(sub).sub
        repl = repl or ''
    for root, _, filenames in os.walk(src):
        for fn in filenames:
            spath = join(root, fn)
            if match(spath):
                if sub is not None:
                    spath = sub(repl, spath)
                dpath = join(dst, relpath(spath, src))
                os.renames(spath, dpath)


def fix_shebang(fpath, replacement=None):
    """Normalize file's shebang.

    :param replacement: new shebang command (path to interpreter and options)
    """
    try:
        interpreter = Interpreter.from_file(fpath)
    except Exception as err:
        log.debug('fix_shebang (%s): %s', fpath, err)
        return None

    if not replacement and interpreter.version == '2':
        # we'll drop /usr/bin/python symlink from python package at some point
        replacement = '/usr/bin/python2'
        if interpreter.debug:
            replacement += '-dbg'
    elif not replacement and interpreter.path != '/usr/bin/':  # f.e. /usr/local/* or */bin/env
        interpreter.path = '/usr/bin'
        replacement = repr(interpreter)
    if replacement:
        log.info('replacing shebang in %s', fpath)
        try:
            with open(fpath, 'rb') as fp:
                fcontent = fp.readlines()
        except IOError:
            log.error('cannot open %s', fpath)
            return False
        # do not catch IOError here, the file is zeroed at this stage so it's
        # better to fail
        with open(fpath, 'wb') as fp:
            fp.write(("#! %s\n" % replacement).encode('utf-8'))
            fp.writelines(fcontent[1:])
    return True


def so2pyver(fpath):
    """Return libpython version file is linked to or None.

    :rtype: tuple
    :returns: Python version
    """

    cmd = "readelf -Wd '%s'" % fpath
    with Popen(cmd, stdout=PIPE, shell=True) as process:
        match = SHAREDLIB_RE.search(str(process.stdout.read(), encoding="UTF-8"))
        if match:
            return Version(match.groups()[0])


def clean_egg_name(name):
    """Remove Python version and platform name from Egg files/dirs.

    >>> clean_egg_name('python_pipeline-0.1.3_py3k-py3.1.egg-info')
    'python_pipeline-0.1.3_py3k.egg-info'
    >>> clean_egg_name('Foo-1.2-py2.7-linux-x86_64.egg-info')
    'Foo-1.2.egg-info'
    """
    match = EGGnPTH_RE.match(name)
    if match and match.group(2) is not None:
        return ''.join(match.group(1, 3))
    return name


def parse_ns(fpaths, other=None):
    """Parse namespace_packages.txt files."""
    result = set(other or [])
    for fpath in fpaths:
        with open(fpath, 'r', encoding='utf-8') as fp:
            for line in fp:
                if line:
                    result.add(line.strip())
    return result


def remove_ns(interpreter, package, namespaces, versions):
    """Remove empty __init__.py files for requested namespaces."""
    if not isinstance(namespaces, set):
        namespaces = set(namespaces)
    keep = set()
    for ns in namespaces:
        for version in versions:
            fpath = join(interpreter.sitedir(package, version), *ns.split('.'))
            fpath = join(fpath, '__init__.py')
            if not exists(fpath):
                continue
            if getsize(fpath) != 0:
                log.warning('file not empty, cannot share %s namespace', ns)
                keep.add(ns)
                break

    # return a set of namespaces that should be handled by pycompile/pyclean
    result = namespaces - keep

    # remove empty __init__.py files, if available
    for ns in result:
        for version in versions:
            dpath = join(interpreter.sitedir(package, version), *ns.split('.'))
            fpath = join(dpath, '__init__.py')
            if exists(fpath):
                os.remove(fpath)
                if not os.listdir(dpath):
                    os.rmdir(dpath)
        # clean pyshared dir as well
        dpath = join('debian', package, 'usr/share/pyshared', *ns.split('.'))
        fpath = join(dpath, '__init__.py')
        if exists(fpath):
            os.remove(fpath)
            if not os.listdir(dpath):
                os.rmdir(dpath)
    return result


def execute(command, cwd=None, env=None, log_output=None, shell=True):
    """Execute external shell command.

    :param cdw: current working directory
    :param env: environment
    :param log_output:
        * opened log file or path to this file, or
        * None if output should be included in the returned dict, or
        * False if output should be redirected to stdout/stderr
    """
    args = {'shell': shell, 'cwd': cwd, 'env': env}
    close = False
    if log_output is False:
        pass
    elif log_output is None:
        args.update(stdout=PIPE, stderr=PIPE)
    elif log_output:
        if isinstance(log_output, str):
            close = True
            # pylint: disable=consider-using-with
            log_output = open(log_output, 'a', encoding='utf-8')
        log_output.write('\n# command executed on {}'.format(datetime.now().isoformat()))
        log_output.write('\n$ {}\n'.format(command))
        log_output.flush()
        args.update(stdout=log_output, stderr=log_output)

    log.debug('invoking: %s', command)
    with Popen(command, **args) as process:
        stdout, stderr = process.communicate()
        if close:
            log_output.close()
        return {
            "returncode": process.returncode,
            "stdout": stdout is not None and str(stdout, 'utf-8'),
            "stderr": stderr is not None and str(stderr, 'utf-8'),
        }


class memoize:
    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args, **kwargs):
        key = dumps((args, kwargs))
        if key not in self.cache:
            self.cache[key] = self.func(*args, **kwargs)
        return self.cache[key]


@memoize
def dpkg_architecture():
    """Parse dpkg-architecture output"""
    arch_data = {}
    if exists('/usr/bin/dpkg-architecture'):
        res = execute('/usr/bin/dpkg-architecture')
        for line in res['stdout'].splitlines():
            key, value = line.strip().split('=', 1)
            arch_data[key] = value
    return arch_data


def pyinstall(interpreter, package, vrange):
    """Install local files listed in pkg.pyinstall files as public modules."""
    srcfpath = "./debian/%s.pyinstall" % package
    if not exists(srcfpath):
        return
    impl = interpreter.impl
    versions = get_requested_versions(impl, vrange)

    with open(srcfpath, encoding='utf-8') as fp:
        for line in fp:
            if not line.strip() or line.startswith('#'):
                continue
            details = INSTALL_RE.match(line)
            if not details:
                raise ValueError("unrecognized line: %s" % line)
            details = details.groupdict()
            if details['module']:
                details['module'] = details['module'].replace('.', '/')
            myvers = versions & get_requested_versions(impl, details['vrange'])
            if not myvers:
                log.debug('%s.pyinstall: no matching versions for line %s',
                        package, line)
                continue
            files = glob(details['pattern'])
            if not files:
                raise ValueError("missing file(s): %s" % details['pattern'])
            for fpath in files:
                fpath = fpath.lstrip('/.')
                if details['module']:
                    dstname = join(details['module'], split(fpath)[1])
                elif fpath.startswith('debian/'):
                    dstname = fpath[7:]
                else:
                    dstname = fpath
                for version in myvers:
                    dstfpath = join(interpreter.sitedir(package, version), dstname)
                    dstdir = split(dstfpath)[0]
                    if not exists(dstdir):
                        os.makedirs(dstdir)
                    if exists(dstfpath):
                        os.remove(dstfpath)
                    os.link(fpath, dstfpath)


def pyremove(interpreter, package, vrange):
    """Remove public modules listed in pkg.pyremove file."""
    srcfpath = "./debian/%s.pyremove" % package
    if not exists(srcfpath):
        return
    impl = interpreter.impl
    versions = get_requested_versions(impl, vrange)

    with open(srcfpath, encoding='utf-8') as fp:
        for line in fp:
            if not line.strip() or line.startswith('#'):
                continue
            details = REMOVE_RE.match(line)
            if not details:
                raise ValueError("unrecognized line: %s: %s" % (package, line))
            details = details.groupdict()
            myvers = versions & get_requested_versions(impl, details['vrange'])
            if not myvers:
                log.debug('%s.pyremove: no matching versions for line %s',
                        package, line)
            for version in myvers:
                site_dirs = interpreter.old_sitedirs(package, version)
                site_dirs.append(interpreter.sitedir(package, version))
                for sdir in site_dirs:
                    files = glob(sdir + '/' + details['pattern'])
                    for fpath in files:
                        if isdir(fpath):
                            rmtree(fpath)
                        else:
                            os.remove(fpath)

from dhpython.interpreter import Interpreter
from dhpython.version import Version, get_requested_versions, RANGE_PATTERN
INSTALL_RE = re.compile(r"""
    (?P<pattern>.+?)  # file pattern
    (?:\s+  # optional Python module name:
    (?P<module>[A-Za-z][A-Za-z0-9_.]*)?
    )?
    \s*  # optional version range:
    (?P<vrange>%s)?$
""" % RANGE_PATTERN, re.VERBOSE)
REMOVE_RE = re.compile(r"""
    (?P<pattern>.+?)  # file pattern
    \s*  # optional version range:
    (?P<vrange>%s)?$
""" % RANGE_PATTERN, re.VERBOSE)



import sys
import unittest
from os import environ
from os.path import exists
from dhpython.interpreter import Interpreter


class TestInterpreter(unittest.TestCase):
    def setUp(self):
        self._triplet = environ.get('DEB_HOST_MULTIARCH')
        environ['DEB_HOST_MULTIARCH'] = 'MYARCH'

    def tearDown(self):
        if self._triplet:
            environ['DEB_HOST_MULTIARCH'] = self._triplet
        else:
            del environ['DEB_HOST_MULTIARCH']

    @unittest.skipUnless(exists('/usr/bin/python3.1'), 'python3.1 is not installed')
    def test_python31(self):
        i = Interpreter('python3.1')
        self.assertEqual(i.soabi(), '')
        self.assertIsNone(i.check_extname('foo.so'))
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertIsNone(i.check_extname('foo/bar/bazmodule.so'))

    @unittest.skipUnless(exists('/usr/bin/python3.1-dbg'), 'python3.1-dbg is not installed')
    def test_python31dbg(self):
        i = Interpreter('python3.1-dbg')
        self.assertEqual(i.soabi(), '')
        self.assertIsNone(i.check_extname('foo.so'))
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertIsNone(i.check_extname('foo/bar/bazmodule.so'))

    @unittest.skipUnless(exists('/usr/bin/python3.2'), 'python3.2 is not installed')
    def test_python32(self):
        i = Interpreter('python3.2')
        self.assertEqual(i.soabi(), 'cpython-32mu')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-32mu.so')
        self.assertIsNone(i.check_extname('foo.cpython-33m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-32mu-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/bazmodule.cpython-32mu.so')

    @unittest.skipUnless(exists('/usr/bin/python3.2-dbg'), 'python3.2-dbg is not installed')
    def test_python32dbg(self):
        i = Interpreter('python3.2-dbg')
        self.assertEqual(i.soabi(), 'cpython-32dmu')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-32dmu.so')
        self.assertIsNone(i.check_extname('foo.cpython-33m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-32dmu-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/bazmodule.cpython-32dmu.so')

    @unittest.skipUnless(exists('/usr/bin/python3.4'), 'python3.4 is not installed')
    def test_python34(self):
        i = Interpreter('python3.4')
        self.assertEqual(i.soabi(), 'cpython-34m')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-34m-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-34m-OTHER.so'))  # different architecture
        self.assertEqual(i.check_extname('foo.cpython-34m.so'), r'foo.cpython-34m-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-34m-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.4-dbg'), 'python3.4-dbg is not installed')
    def test_python34dbg(self):
        i = Interpreter('python3.4-dbg')
        self.assertEqual(i.soabi(), 'cpython-34dm')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-34dm-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-34m-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-34dm-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.5'), 'python3.5 is not installed')
    def test_python35(self):
        i = Interpreter('python3.5')
        self.assertEqual(i.soabi(), 'cpython-35m')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-35m-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-35m-OTHER.so'))  # different architecture
        self.assertEqual(i.check_extname('foo.cpython-35m.so'), r'foo.cpython-35m-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-35m-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.5-dbg'), 'python3.5-dbg is not installed')
    def test_python35dbg(self):
        i = Interpreter('python3.5-dbg')
        self.assertEqual(i.soabi(), 'cpython-35dm')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-35dm-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-35m-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-35dm-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.6'), 'python3.6 is not installed')
    def test_python36(self):
        i = Interpreter('python3.6')
        self.assertEqual(i.soabi(), 'cpython-36m')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-36m-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-36m-OTHER.so'))  # different architecture
        self.assertEqual(i.check_extname('foo.cpython-36m.so'), r'foo.cpython-36m-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-36m-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.6-dbg'), 'python3.6-dbg is not installed')
    def test_python36dbg(self):
        i = Interpreter('python3.6-dbg')
        self.assertEqual(i.soabi(), 'cpython-36dm')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-36dm-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-36m-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-36dm-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.7'), 'python3.7 is not installed')
    def test_python37(self):
        i = Interpreter('python3.7')
        self.assertEqual(i.soabi(), 'cpython-37m')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-37m-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-37m-OTHER.so'))  # different architecture
        self.assertEqual(i.check_extname('foo.cpython-37m.so'), r'foo.cpython-37m-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-37m-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.7-dbg'), 'python3.7-dbg is not installed')
    def test_python37dbg(self):
        i = Interpreter('python3.7-dbg')
        self.assertEqual(i.soabi(), 'cpython-37dm')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-37dm-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-37m-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-37dm-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.8'), 'python3.8 is not installed')
    def test_python38(self):
        i = Interpreter('python3.8')
        self.assertEqual(i.soabi(), 'cpython-38')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-38-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-38-OTHER.so'))  # different architecture
        self.assertEqual(i.check_extname('foo.cpython-38.so'), r'foo.cpython-38-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-38-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.8-dbg'), 'python3.8-dbg is not installed')
    def test_python38dbg(self):
        i = Interpreter('python3.8-dbg')
        self.assertEqual(i.soabi(), 'cpython-38d')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-38d-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-38-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-38d-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.9'), 'python3.9 is not installed')
    def test_python39(self):
        i = Interpreter('python3.9')
        self.assertEqual(i.soabi(), 'cpython-39')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-39-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-39-OTHER.so'))  # different architecture
        self.assertEqual(i.check_extname('foo.cpython-39.so'), r'foo.cpython-39-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-39-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.9-dbg'), 'python3.9-dbg is not installed')
    def test_python39dbg(self):
        i = Interpreter('python3.9-dbg')
        self.assertEqual(i.soabi(), 'cpython-39d')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-39d-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-39-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-39d-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.10'), 'python3.10 is not installed')
    def test_python310(self):
        i = Interpreter('python3.10')
        self.assertEqual(i.soabi(), 'cpython-310')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-310-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-310-OTHER.so'))  # different architecture
        self.assertEqual(i.check_extname('foo.cpython-310.so'), r'foo.cpython-310-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-310-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.10-dbg'), 'python3.10-dbg is not installed')
    def test_python310dbg(self):
        i = Interpreter('python3.10-dbg')
        self.assertEqual(i.soabi(), 'cpython-310d')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-310d-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-310-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-310d-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.11'), 'python3.11 is not installed')
    def test_python311(self):
        i = Interpreter('python3.11')
        self.assertEqual(i.soabi(), 'cpython-311')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-311-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-311-OTHER.so'))  # different architecture
        self.assertEqual(i.check_extname('foo.cpython-311.so'), r'foo.cpython-311-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-311-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.11-dbg'), 'python3.11-dbg is not installed')
    def test_python311dbg(self):
        i = Interpreter('python3.11-dbg')
        self.assertEqual(i.soabi(), 'cpython-311d')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-311d-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-311-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-311d-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.12'), 'python3.12 is not installed')
    def test_python312(self):
        i = Interpreter('python3.12')
        self.assertEqual(i.soabi(), 'cpython-312')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-312-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-312-OTHER.so'))  # different architecture
        self.assertEqual(i.check_extname('foo.cpython-312.so'), r'foo.cpython-312-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-312-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3.12-dbg'), 'python3.12-dbg is not installed')
    def test_python312dbg(self):
        i = Interpreter('python3.12-dbg')
        self.assertEqual(i.soabi(), 'cpython-312d')
        self.assertEqual(i.check_extname('foo.so'), r'foo.cpython-312d-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname('foo.cpython-312-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), r'foo/bar/baz.cpython-312d-MYARCH.so')

    def test_python3(self):
        i = Interpreter('python{}.{}'.format(*sys.version_info[:2]))
        pyver = '{}{}'.format(*sys.version_info[:2])
        self.assertEqual(i.soabi(), 'cpython-' + pyver)
        self.assertEqual(i.check_extname('foo.so'), rf'foo.cpython-{pyver}-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname(f'foo.cpython-{pyver}-OTHER.so'))  # different architecture
        self.assertEqual(i.check_extname(f'foo.cpython-{pyver}.so'), rf'foo.cpython-{pyver}-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), rf'foo/bar/baz.cpython-{pyver}-MYARCH.so')

    @unittest.skipUnless(exists('/usr/bin/python3-dbg'), 'python3-dbg is not installed')
    def test_python3dbg(self):
        i = Interpreter('python{}.{}-dbg'.format(*sys.version_info[:2]))
        pyver = '{}{}'.format(*sys.version_info[:2])
        self.assertEqual(i.soabi(), f'cpython-{pyver}d')
        self.assertEqual(i.check_extname('foo.so'), rf'foo.cpython-{pyver}d-MYARCH.so')
        self.assertIsNone(i.check_extname('foo.cpython-32m.so'))  # different version
        self.assertIsNone(i.check_extname(f'foo.cpython-{pyver}-OTHER.so'))  # different architecture
        self.assertIsNone(i.check_extname('foo.abi3.so'))
        self.assertEqual(i.check_extname('foo/bar/bazmodule.so'), rf'foo/bar/baz.cpython-{pyver}d-MYARCH.so')

    def test_bare_module(self):
        i = Interpreter('python{}.{}'.format(*sys.version_info[:2]))
        pyver = '{}{}'.format(*sys.version_info[:2])
        self.assertEqual(i.check_extname('module.so'), rf'module.cpython-{pyver}-MYARCH.so')
        self.assertEqual(i.check_extname('_module.so'), rf'_module.cpython-{pyver}-MYARCH.so')


if __name__ == '__main__':
    unittest.main()



# Copyright © 2010-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import logging
import re
from os import makedirs, chmod, environ
from os.path import basename, exists, join, dirname
from sys import argv
from dhpython import DEPENDS_SUBSTVARS, PKG_NAME_TPLS, RT_LOCATIONS, RT_TPLS

log = logging.getLogger('dhpython')
parse_dep = re.compile(r'''[,\s]*
    (?P<name>[^\s:]+)(?::any)?
    \s*
    \(?(?P<version>([>=<]{2,}|=)\s*[^\)]+)?\)?
    \s*
    (?:\[(?P<arch>[^\]]+)\])?
    ''', re.VERBOSE).match


def build_options(**options):
    """Build an Options object from kw options"""
    default_options = {
        'arch': None,
        'package': [],
        'no_package': [],
        'write_log': False,
        'remaining_packages': False,
    }
    built_options = default_options
    built_options.update(options)
    return type('Options', (object,), built_options)


class DebHelper:
    """Reinvents the wheel / some dh functionality (Perl is ugly ;-P)"""

    def __init__(self, options, impl='cpython3'):
        self.options = options
        self.packages = {}
        self.build_depends = {}
        self.python_version = None
        self.impl = impl
        self.command = {
            'cpython3': 'dh_python3',
        }[impl]
        skip_tpl = set()
        for name, tpls in PKG_NAME_TPLS.items():
            if name != impl:
                skip_tpl.update(tpls)
        skip_tpl = tuple(skip_tpl)
        substvar = DEPENDS_SUBSTVARS[impl]

        pkgs = options.package
        skip_pkgs = options.no_package

        try:
            with open('debian/control', 'r', encoding='utf-8') as fp:
                paragraphs = [{}]
                field = None
                for lineno, line in enumerate(fp, 1):
                    if line.startswith('#'):
                        continue
                    if not line.strip():
                        if paragraphs[-1]:
                            paragraphs.append({})
                            field = None
                        continue
                    if line[0].isspace():  # Continuation
                        paragraphs[-1][field] += line.rstrip()
                        continue
                    if not ':' in line:
                        raise Exception(
                            'Unable to parse line %i in debian/control: %s'
                            % (lineno, line))
                    field, value = line.split(':', 1)
                    field = field.lower()
                    paragraphs[-1][field] = value.strip()
        except IOError:
            raise Exception('cannot find debian/control file')

        # Trailing new lines?
        if not paragraphs[-1]:
            paragraphs.pop()

        if len(paragraphs) < 2:
            raise Exception('Unable to parse debian/control, found less than '
                            '2 paragraphs')

        self.source_name = paragraphs[0]['source']
        if self.impl == 'cpython3' and 'x-python3-version' in paragraphs[0]:
            self.python_version = paragraphs[0]['x-python3-version']
            if len(self.python_version.split(',')) > 2:
                raise ValueError('too many arguments provided for '
                                 'X-Python3-Version: min and max only.')

        build_depends = []
        for field in ('build-depends', 'build-depends-indep',
                      'build-depends-arch'):
            if field in paragraphs[0]:
                build_depends.append(paragraphs[0][field])
        build_depends = ', '.join(build_depends)
        for dep1 in build_depends.split(','):
            for dep2 in dep1.split('|'):
                details = parse_dep(dep2)
                if details:
                    details = details.groupdict()
                    if details['arch']:
                        architectures = details['arch'].split()
                    else:
                        architectures = [None]
                    for arch in architectures:
                        self.build_depends.setdefault(
                            details['name'], {})[arch] = details['version']

        for paragraph_no, paragraph in enumerate(paragraphs[1:], 2):
            if 'package' not in paragraph:
                raise Exception('Unable to parse debian/control, paragraph %i '
                                'missing Package field' % paragraph_no)
            binary_package = paragraph['package']
            if skip_tpl and binary_package.startswith(skip_tpl):
                log.debug('skipping package: %s', binary_package)
                continue
            if pkgs and binary_package not in pkgs:
                continue
            if skip_pkgs and binary_package in skip_pkgs:
                continue
            if (options.remaining_packages and
                    self.has_acted_on_package(binary_package)):
                continue
            pkg = {
                'substvars': {},
                'autoscripts': {},
                'rtupdates': [],
                'arch': paragraph['architecture'],
            }
            if (options.arch is False and pkg['arch'] != 'all' or
                    options.arch is True and pkg['arch'] == 'all'):
                # TODO: check also if arch matches current architecture:
                continue

            if not binary_package.startswith(PKG_NAME_TPLS[impl]):
                # package doesn't have common prefix (python3-)
                # so lets check if Depends/Recommends contains the
                # appropriate substvar
                if (substvar not in paragraph.get('depends', '')
                        and substvar not in paragraph.get('recommends', '')):
                    log.debug('skipping package %s (missing %s in '
                              'Depends/Recommends)',
                              binary_package, substvar)
                    continue
            # Operate on binary_package
            self.packages[binary_package] = pkg

        fp.close()
        log.debug('source=%s, binary packages=%s', self.source_name,
                  list(self.packages.keys()))

    def has_acted_on_package(self, package):
        try:
            with open('debian/{}.debhelper.log'.format(package),
                      encoding='utf-8') as f:
                for line in f:
                    if line.strip() == self.command:
                        return True
        except IOError as e:
            if e.errno != errno.ENOENT:
                raise
        return False

    def addsubstvar(self, package, name, value):
        """debhelper's addsubstvar"""
        self.packages[package]['substvars'].setdefault(name, []).append(value)

    def autoscript(self, package, when, template, args):
        """debhelper's autoscript"""
        self.packages[package]['autoscripts'].setdefault(when, {})\
            .setdefault(template, []).append(args)

    def add_rtupdate(self, package, value):
        self.packages[package]['rtupdates'].append(value)

    def save_autoscripts(self):
        for package, settings in self.packages.items():
            autoscripts = settings.get('autoscripts')
            if not autoscripts:
                continue

            for when, templates in autoscripts.items():
                fn = "debian/%s.%s.debhelper" % (package, when)
                if exists(fn):
                    with open(fn, 'r', encoding='utf-8') as datafile:
                        data = datafile.read()
                else:
                    data = ''

                new_data = ''
                for tpl_name, args in templates.items():
                    for i in args:
                        # try local one first (useful while testing dh_python3)
                        fpath = join(dirname(__file__), '..',
                                     "autoscripts/%s" % tpl_name)
                        if not exists(fpath):
                            fpath = "/usr/share/debhelper/autoscripts/%s" % tpl_name
                        with open(fpath, 'r', encoding='utf-8') as tplfile:
                            tpl = tplfile.read()
                        if self.options.compile_all and args:
                            # TODO: should args be checked to contain dir name?
                            tpl = tpl.replace('-p #PACKAGE#', '')
                        elif settings['arch'] == 'all':
                            tpl = tpl.replace('#PACKAGE#', package)
                        else:
                            arch = environ['DEB_HOST_ARCH']
                            tpl = tpl.replace('#PACKAGE#', '%s:%s' % (package, arch))
                        tpl = tpl.replace('#ARGS#', i)
                        if tpl not in data and tpl not in new_data:
                            new_data += "\n%s" % tpl
                if new_data:
                    data += '\n# Automatically added by {}'.format(basename(argv[0])) +\
                            '{}\n# End automatically added section\n'.format(new_data)
                    with open(fn, 'w', encoding='utf-8') as fp:
                        fp.write(data)

    def save_substvars(self):
        for package, settings in self.packages.items():
            substvars = settings.get('substvars')
            if not substvars:
                continue
            fn = "debian/%s.substvars" % package
            if exists(fn):
                with open(fn, 'r', encoding='utf-8') as datafile:
                    data = datafile.read()
            else:
                data = ''
            for name, values in substvars.items():
                p = data.find("%s=" % name)
                if p > -1:  # parse the line and remove it from data
                    e = data[p:].find('\n')
                    line = data[p + len("%s=" % name):
                                p + e if e > -1 else None]
                    items = [i.strip() for i in line.split(',') if i]
                    if e > -1 and data[p + e:].strip():
                        data = "%s\n%s" % (data[:p], data[p + e:])
                    else:
                        data = data[:p]
                else:
                    items = []
                for j in values:
                    if j not in items:
                        items.append(j)
                if items:
                    if data:
                        data += '\n'
                    data += "%s=%s\n" % (name, ', '.join(items))
            data = data.replace('\n\n', '\n')
            if data:
                with open(fn, 'w', encoding='utf-8') as fp:
                    fp.write(data)

    def save_rtupdate(self):
        for package, settings in self.packages.items():
            pkg_arg = '' if self.options.compile_all else "-p %s" % package
            values = settings.get('rtupdates')
            if not values:
                continue
            d = 'debian/{}/{}'.format(package, RT_LOCATIONS[self.impl])
            if not exists(d):
                makedirs(d)
            fn = "%s/%s.rtupdate" % (d, package)
            if exists(fn):
                with open(fn, 'r', encoding='utf-8') as fp:
                    data = fp.read()
            else:
                data = "#! /bin/sh\nset -e"
            for dname, args in values:
                cmd = RT_TPLS[self.impl].format(pkg_arg=pkg_arg,
                                                dname=dname,
                                                args=args)
                if cmd not in data:
                    data += "\n%s" % cmd
            if data:
                with open(fn, 'w', encoding='utf-8') as fp:
                    fp.write(data)
                chmod(fn, 0o755)

    def save_log(self):
        if not self.options.write_log:
            return
        for package, _ in self.packages.items():
            with open('debian/{}.debhelper.log'.format(package),
                      'a', encoding='utf-8') as f:
                f.write(self.command + '\n')

    def save(self):
        self.save_substvars()
        self.save_autoscripts()
        self.save_rtupdate()
        self.save_log()



# Copyright © 2012-2020 Piotr Ożarowski <piotr@debian.org>
#           © 2020 Scott Kitterman <scott@kitterman.com>
#           © 2021 Stuart Prescott <stuart@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from pathlib import Path
import logging
import os.path as osp
import shutil
import sysconfig
try:
    import tomllib
except ModuleNotFoundError:
    try:
        import tomli as tomllib
    except ModuleNotFoundError:
        # Plugin still works, only needed for autodetection
        pass
try:
    from installer import install
    from installer.destinations import SchemeDictionaryDestination
    from installer.sources import WheelFile
except ModuleNotFoundError:
    SchemeDictionaryDestination = WheelFile = install = None

from dhpython.build.base import Base, shell_command
from dhpython.tools import dpkg_architecture

log = logging.getLogger('dhpython')


class BuildSystem(Base):
    DESCRIPTION = 'Generic PEP517 build system'
    SUPPORTED_INTERPRETERS = {'python3', 'python{version}'}
    REQUIRED_FILES = ['pyproject.toml']
    OPTIONAL_FILES = {}
    CLEAN_FILES = Base.CLEAN_FILES | {'build'}

    def detect(self, context):
        """Return certainty level that this plugin describes the right build
        system

        This method uses cls.{REQUIRED}_FILES (pyroject.toml) only; any
        other PEP517 compliant builder (such as the flit) builder should
        indicate higher specificity than this plugin.

        :return: 0 <= certainty <= 100
        :rtype: int
        """
        result = super().detect(context)
        # Temporarily reduce the threshold while we're in beta
        result -= 20

        if self._build_backend():
            result += 10

        if result > 100:
            return 100
        return result

    def _build_backend(self):
        """Retrieve the build-system from pyproject.toml, where possible"""
        try:
            with open('pyproject.toml', 'rb') as f:
                pyproject = tomllib.load(f)
            return pyproject.get('build-system', {}).get('build-backend')
        except NameError:
            # No toml, no autdetection
            return None
        except FileNotFoundError:
            # Not a PEP517 package
            return None

    def clean(self, context, args):
        super().clean(context, args)
        if osp.exists(args['interpreter'].binary()):
            log.debug("removing '%s' (and everything under it)",
                      args['build_dir'])
            if osp.isdir(args['build_dir']):
                shutil.rmtree(args['build_dir'])
        return 0  # no need to invoke anything

    def configure(self, context, args):
        if install is None:
            raise Exception("PEP517 plugin dependencies are not available. "
                            "Please Build-Depend on pybuild-plugin-pyproject.")
        # No separate configure step
        return 0

    def build(self, context, args):
        self.build_wheel(context, args)
        self.unpack_wheel(context, args)

    def _backend_config_settings(self):
        backend = self._build_backend()
        if backend == "mesonpy":
            arch_data = dpkg_architecture()
            return [
                "compile-args=--verbose",
                # From Debhelper's meson.pm
                "setup-args=--wrap-mode=nodownload",
                "setup-args=--prefix=/usr",
                "setup-args=--sysconfdir=/etc",
                "setup-args=--localstatedir=/var",
                "setup-args=--libdir=lib/" + arch_data["DEB_HOST_MULTIARCH"],
            ]
        return []

    @shell_command
    def build_wheel(self, context, args):
        """ build a wheel using the PEP517 builder defined by upstream """
        log.info('Building wheel for %s with "build" module',
                 args['interpreter'])
        config_settings = self._backend_config_settings()
        context['ENV']['FLIT_NO_NETWORK'] = '1'
        context['ENV']['HOME'] = args['home_dir']
        return ('{interpreter} -m build '
                '--skip-dependency-check --no-isolation --wheel '
                '--outdir ' + args['home_dir'] + ' ' +
                ' '.join(('--config-setting ' + setting)
                         for setting in config_settings) +
                ' {args}'
               )

    def unpack_wheel(self, context, args):
        """ unpack the wheel into pybuild's normal  """
        log.info('Unpacking wheel built for %s with "installer" module',
                 args['interpreter'])
        extras = {}
        for extra in ('scripts', 'data'):
            path = Path(args["home_dir"]) / extra
            if osp.exists(path):
                log.warning('%s directory already exists, skipping unpack. '
                            'Is the Python package being built twice?',
                            extra.title())
                return
            extras[extra] = path
        destination = SchemeDictionaryDestination(
            {
                'platlib': args['build_dir'],
                'purelib': args['build_dir'],
                'scripts': extras['scripts'],
                'data': extras['data'],
            },
            interpreter=args['interpreter'].binary_dv,
            script_kind='posix',
        )

        wheel = Path(self.built_wheel(context, args))
        if wheel.name.startswith('UNKNOWN'):
            raise Exception(f'UNKNOWN wheel found: {wheel.name}. Does '
                            'pyproject.toml specify a build-backend?')
        with WheelFile.open(wheel) as source:
            install(
                source=source,
                destination=destination,
                additional_metadata={},
            )

    def install(self, context, args):
        log.info('Copying package built for %s to destdir',
                 args['interpreter'])
        try:
            paths = sysconfig.get_paths(scheme='deb_system')
        except KeyError:
            # Debian hasn't patched sysconfig schemes until 3.10
            # TODO: Introduce a version check once sysconfig is patched.
            paths = sysconfig.get_paths(scheme='posix_prefix')

        # start by copying the data and scripts
        for extra in ('data', 'scripts'):
            src_dir = Path(args['home_dir']) / extra
            if not src_dir.exists():
                continue
            target_dir = args['destdir'] + paths[extra]
            log.debug('Copying %s directory contents from %s -> %s',
                      extra, src_dir, target_dir)
            shutil.copytree(
                src_dir,
                target_dir,
                dirs_exist_ok=True,
            )

        # then copy the modules
        module_dir = args['build_dir']
        target_dir = args['destdir'] + args['install_dir']
        log.debug('Copying module contents from %s -> %s',
                  module_dir, target_dir)
        shutil.copytree(
            module_dir,
            target_dir,
            dirs_exist_ok=True,
        )

    @shell_command
    def test(self, context, args):
        scripts = Path(args["home_dir"]) / 'scripts'
        if scripts.exists():
            context['ENV']['PATH'] = f"{scripts}:{context['ENV']['PATH']}"
        context['ENV']['HOME'] = args['home_dir']
        return super().test(context, args)



# vim: et ts=4 sw=4
# Copyright © 2021 Antonio Terceiro <terceiro@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from dhpython.build.base import Base, shell_command


class BuildSystem(Base):
    """
    autopkgtest test runner. All the methods that are not related to
    testing (configure, build, clean etc) are not implemented by this class, on
    purpose.
    """

    NAME = "autopkgtest"
    DESCRIPTION = 'autopkgtest test runner'
    SUPPORTED_INTERPRETERS = {'python3', 'python{version}'}

    @shell_command
    def test(self, context, args):
        return super().test(context, args)
